@@ -57,10 +57,6 @@ public:
     ) override;
   friend struct SubWriteApplied;
   friend struct SubWriteCommitted;
-  void sub_write_applied(
-    ceph_tid_t tid,
-    eversion_t version,
-    const ZTracer::Trace &trace);
   void sub_write_committed(
     ceph_tid_t tid,
     eversion_t version,
@@ -70,8 +66,7 @@ public:
     pg_shard_t from,
     OpRequestRef msg,
     ECSubWrite &op,
-    const ZTracer::Trace &trace,
-    Context *on_local_applied_sync = 0
+    const ZTracer::Trace &trace
     );
   void handle_sub_read(
     pg_shard_t from,
@@ -97,8 +92,6 @@ public:
   void on_change() override;
   void clear_recovery_state() override;
 
-  void on_flushed() override;
-
   void dump_recovery_info(Formatter *f) const override;
 
   void call_write_ordered(std::function<void(void)> &&cb) override;
@@ -112,8 +105,6 @@ public:
     const eversion_t &roll_forward_to,
     const vector<pg_log_entry_t> &log_entries,
     boost::optional<pg_hit_set_history_t> &hset_history,
-    Context *on_local_applied_sync,
-    Context *on_all_applied,
     Context *on_all_commit,
     ceph_tid_t tid,
     osd_reqid_t reqid,
@@ -140,7 +131,7 @@ public:
    * maintain a queue of in progress reads (@see in_progress_client_reads)
    * to ensure that we always call the completion callback in order.
    *
-   * Another subtely is that while we may read a degraded object, we will
+   * Another subtly is that while we may read a degraded object, we will
    * still only perform a client read from shards in the acting set.  This
    * ensures that we won't ever have to restart a client initiated read in
    * check_recovery_sources.
@@ -164,9 +155,9 @@ public:
       const hobject_t &hoid,
       int err,
       extent_map &&buffers) {
-      assert(objects_to_read);
+      ceph_assert(objects_to_read);
       --objects_to_read;
-      assert(!results.count(hoid));
+      ceph_assert(!results.count(hoid));
       results.emplace(hoid, make_pair(err, std::move(buffers)));
     }
     bool is_complete() const {
@@ -213,7 +204,7 @@ public:
 private:
   friend struct ECRecoveryHandle;
   uint64_t get_recovery_chunk_size() const {
-    return ROUND_UP_TO(cct->_conf->osd_recovery_max_chunk,
+    return round_up_to(cct->_conf->osd_recovery_max_chunk,
 			sinfo.get_stripe_width());
   }
 
@@ -315,7 +306,8 @@ private:
     RecoveryMessages *m);
   void handle_recovery_push(
     const PushOp &op,
-    RecoveryMessages *m);
+    RecoveryMessages *m,
+    bool is_repair);
   void handle_recovery_push_reply(
     const PushReplyOp &op,
     pg_shard_t from,
@@ -360,12 +352,12 @@ public:
   };
   struct read_request_t {
     const list<boost::tuple<uint64_t, uint64_t, uint32_t> > to_read;
-    const set<pg_shard_t> need;
+    const map<pg_shard_t, vector<pair<int, int>>> need;
     const bool want_attrs;
     GenContext<pair<RecoveryMessages *, read_result_t& > &> *cb;
     read_request_t(
       const list<boost::tuple<uint64_t, uint64_t, uint32_t> > &to_read,
-      const set<pg_shard_t> &need,
+      const map<pg_shard_t, vector<pair<int, int>>> &need,
       bool want_attrs,
       GenContext<pair<RecoveryMessages *, read_result_t& > &> *cb)
       : to_read(to_read), need(need), want_attrs(want_attrs),
@@ -459,7 +451,7 @@ public:
    * on the writing list.
    */
   struct Op : boost::intrusive::list_base_hook<> {
-    /// From submit_transaction caller, decribes operation
+    /// From submit_transaction caller, describes operation
     hobject_t hoid;
     object_stat_sum_t delta_stats;
     eversion_t version;
@@ -487,7 +479,7 @@ public:
     bool invalidates_cache() const { return plan.invalidates_cache; }
 
     // must be true if requires_rmw(), must be false if invalidates_cache()
-    bool using_cache = false;
+    bool using_cache = true;
 
     /// In progress read state;
     map<hobject_t,extent_set> pending_read; // subset already being read
@@ -497,8 +489,11 @@ public:
       return !remote_read.empty() && remote_read_result.empty();
     }
 
-    /// In progress write state
+    /// In progress write state.
     set<pg_shard_t> pending_commit;
+    // we need pending_apply for pre-mimic peers so that we don't issue a
+    // read on a remote shard before it has applied a previous write.  We can
+    // remove this after nautilus.
     set<pg_shard_t> pending_apply;
     bool write_in_progress() const {
       return !pending_commit.empty() || !pending_apply.empty();
@@ -511,12 +506,8 @@ public:
     ExtentCache::write_pin pin;
 
     /// Callbacks
-    Context *on_local_applied_sync = nullptr;
-    Context *on_all_applied = nullptr;
     Context *on_all_commit = nullptr;
     ~Op() {
-      delete on_local_applied_sync;
-      delete on_all_applied;
       delete on_all_commit;
     }
   };
@@ -542,7 +533,7 @@ public:
    * reused).  Next, factor this part into a class and maintain one per
    * ordering token.  Next, fixup PrimaryLogPG's repop queue to be
    * partitioned by ordering token.  Finally, refactor the op pipeline
-   * so that the log entries passed into submit_tranaction aren't
+   * so that the log entries passed into submit_transaction aren't
    * versioned.  We can't assign versions to them until we actually
    * submit the operation.  That's probably going to be the hard part.
    */
@@ -585,7 +576,7 @@ public:
   /**
    * ECRecPred
    *
-   * Determines the whether _have is suffient to recover an object
+   * Determines the whether _have is sufficient to recover an object
    */
   class ECRecPred : public IsPGRecoverablePredicate {
     set<int> want;
@@ -603,18 +594,25 @@ public:
 	   ++i) {
 	have.insert(i->shard);
       }
-      set<int> min;
+      map<int, vector<pair<int, int>>> min;
       return ec_impl->minimum_to_decode(want, have, &min) == 0;
     }
   };
-  IsPGRecoverablePredicate *get_is_recoverable_predicate() override {
+  IsPGRecoverablePredicate *get_is_recoverable_predicate() const override {
     return new ECRecPred(ec_impl);
   }
 
+  int get_ec_data_chunk_count() const override {
+    return ec_impl->get_data_chunk_count();
+  }
+  int get_ec_stripe_chunk_size() const override {
+    return sinfo.get_chunk_size();
+  }
+
   /**
    * ECReadPred
    *
-   * Determines the whether _have is suffient to read an object
+   * Determines the whether _have is sufficient to read an object
    */
   class ECReadPred : public IsPGReadablePredicate {
     pg_shard_t whoami;
@@ -627,7 +625,7 @@ public:
       return _have.count(whoami) && rec_pred(_have);
     }
   };
-  IsPGReadablePredicate *get_is_readable_predicate() override {
+  IsPGReadablePredicate *get_is_readable_predicate() const override {
     return new ECReadPred(get_parent()->whoami_shard(), ec_impl);
   }
 
@@ -641,7 +639,7 @@ public:
 public:
   ECBackend(
     PGBackend::Listener *pg,
-    coll_t coll,
+    const coll_t &coll,
     ObjectStore::CollectionHandle &ch,
     ObjectStore *store,
     CephContext *cct,
@@ -654,7 +652,7 @@ public:
     const set<int> &want,      ///< [in] desired shards
     bool for_recovery,         ///< [in] true if we may use non-acting replicas
     bool do_redundant_reads,   ///< [in] true if we want to issue redundant reads to reduce latency
-    set<pg_shard_t> *to_read   ///< [out] shards to read
+    map<pg_shard_t, vector<pair<int, int>>> *to_read   ///< [out] shards, corresponding subchunks to read
     ); ///< @return error code, 0 on success
 
   int get_remaining_shards(
@@ -662,7 +660,7 @@ public:
     const set<int> &avail,
     const set<int> &want,
     const read_result_t &result,
-    set<pg_shard_t> *to_read,
+    map<pg_shard_t, vector<pair<int, int>>> *to_read,
     bool for_recovery);
 
   int objects_get_attrs(
@@ -674,7 +672,6 @@ public:
     uint64_t old_size,
     ObjectStore::Transaction *t) override;
 
-  bool scrub_supported() override { return true; }
   bool auto_repair_supported() const override { return true; }
 
   int be_deep_scrub(
